import pprint
from urllib import request
from urllib.request import Request, urlopen

import click
import anyconfig
import os
import flatdict

import msgspec
from msgspec.json import decode

from deepdiff import DeepDiff

from yandex_quasar_cli import settings
from yandex_quasar_cli.quasar_types.quasar_device_action import *
from yandex_quasar_cli.quasar_types.quasar_primary import QuasarResponse, DeviceState, Status


def quasar_save_snapshot():
    """Save quasar state as snapshot"""
    config_check()
    req = Request("{}/user/info".format(settings.quasar_conf["base_url"]))
    req.method = "GET"
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    data = msgspec.json.decode(urlopen(req).read())
    settings.quasar_conf["snapshot"] = data
    anyconfig.dump(settings.quasar_conf, settings.CONFIG_FILE, ac_parser=settings.CONFIG_FORMAT)


def exclude_device_capabilities_properties_callback(obj, path):
    return True if "capabilities" in path or "properties" in path or "request_id" in path else False


def exclude_all_but_external_ids_callback(obj, path):
    return False if "external_id" in path else True


def quasar_compare_snapshot(full, no_props_caps, external_ids, ids):
    """Compare quasar state with saved snapshot"""
    config_check()
    req = Request("{}/user/info".format(settings.quasar_conf["base_url"]))
    req.method = "GET"
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))

    compare_entity = "devices"

    data_current = msgspec.json.decode(urlopen(req).read())[compare_entity]
    data_snapshot = settings.quasar_conf["snapshot"][compare_entity]

    ddiff = None

    if full:
        ddiff = DeepDiff(data_snapshot, data_current, ignore_order=True, ignore_numeric_type_changes=True,
                         ignore_private_variables=True, group_by="external_id")
        click.echo(pprint.pprint(ddiff))

    elif no_props_caps:
        ddiff = DeepDiff(data_snapshot, data_current, ignore_order=True, ignore_numeric_type_changes=True,
                         ignore_private_variables=True,
                         exclude_obj_callback=exclude_device_capabilities_properties_callback, group_by="external_id")
        click.echo(pprint.pprint(ddiff))
    elif external_ids or ids:
        #flat_data_snapshot = flatdict.FlatDict(data_snapshot["devices"], delimiter='.')
        #flat_data_current = flatdict.FlatDict(data_current["devices"], delimiter='.')

        ddiff = DeepDiff(data_snapshot, data_current, ignore_order=True, ignore_numeric_type_changes=True,
                         ignore_private_variables=True,
                         group_by="external_id")
        if "dictionary_item_added" in ddiff.keys():
            for added in ddiff["dictionary_item_added"]:
                if added.count("[") == 1:
                    external_id = added.replace("root['", "").replace("']", "")
                    if external_ids:
                        click.echo("+{}".format(external_id))
                    elif ids:
                        click.echo("+{} # {}".format(external_id, quasar_get_device_id_by_external_id(external_id)))
        if "dictionary_item_removed" in ddiff.keys():
            for removed in ddiff["dictionary_item_removed"]:
                if removed.count("[") == 1:
                    external_id = removed.replace("root['", "").replace("']", "")
                    if external_ids:
                        click.echo("-{}".format(external_id))
                    elif ids:
                        click.echo("-{} # {}".format(external_id, quasar_get_device_id_by_external_id(external_id)))
    else:
        flat_data_snapshot = flatdict.FlatterDict(data_snapshot, delimiter='.')
        flat_data_current = flatdict.FlatterDict(data_current, delimiter='.')
        ddiff = DeepDiff(flat_data_snapshot, flat_data_current, ignore_order=True, ignore_numeric_type_changes=True,
                         ignore_private_variables=True)
        click.echo(pprint.pprint(ddiff))


def quasar_get_full_info(object_type, tsv):
    """Smart home info"""
    config_check()
    req = Request("{}/user/info".format(settings.quasar_conf["base_url"]))
    req.method = "GET"
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    data = decode(urlopen(req).read(), type=QuasarResponse)
    if object_type == "devices":
        obj = data.devices
    elif object_type == "rooms":
        obj = data.rooms
    elif object_type == "households":
        obj = data.households
    elif object_type == "groups":
        obj = data.groups
    elif object_type == "scenarios":
        obj = data.scenarios
    else:
        obj = data.devices

    if tsv:
        for record in obj:
            record.print_tsv()
    else:
        for record in obj:
            record.print()


def quasar_get_info(device_id, tsv):
    """Device info"""
    config_check()
    req = Request("{}/devices/{}".format(settings.quasar_conf["base_url"], device_id))
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    req.method = "GET"
    obj = decode(urlopen(req).read(), type=DeviceState)
    if tsv:
        obj.print_tsv()
    else:
        obj.print()


def quasar_delete(ext, device_id):
    config_check()
    del_id = device_id
    if ext:
        del_id = quasar_get_device_id_by_external_id(device_id)

    req = Request("{}/devices/{}".format(settings.quasar_conf["base_url"], del_id))
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    req.method = "DELETE"
    data = decode(urlopen(req).read(), type=Status)
    click.echo("Deleting object {}. Result = {}".format(device_id, data.status))


def quasar_on_off(device_id, state):
    config_check()
    capability_state = CapabilityStateObject(instance="on", value=state)
    capability = CapabilityObject(type="devices.capabilities.on_off", state=capability_state)
    actions = list()
    actions.append(capability)
    device_action = DeviceActionObject(id=device_id, actions=actions)
    devices = list()
    devices.append(device_action)

    device_action_request = DeviceActionRequest(devices)
    json = msgspec.json.encode(device_action_request)
    req = Request("{}/devices/actions".format(settings.quasar_conf["base_url"]), data=json)
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    req.add_header('Content-Type', 'application/json')
    req.method = "POST"
    resp = request.urlopen(req)
    click.echo(resp)


def oauth_help():
    click.echo("How to get new OAuth key")
    click.echo("1. Go to https://oauth.yandex.ru/client/new")
    click.echo("2. Enter service name")
    click.echo("3. Select Web-service platform")
    click.echo("4. Enter https://oauth.yandex.ru/verification_code as redirect URI")
    click.echo("5. Select iot:view and iot:control data access types")
    click.echo("6. Enter email")
    click.echo("7. Click on Create application button")
    click.echo("8. Get ClientID from application info page")
    click.echo("9. Go to page https://oauth.yandex.ru/authorize?response_type=token&client_id=<client_id>")
    click.echo("10. Save oauth key using yandex-quasar-cli save-key command")


def oauth_save_key(key):
    settings.quasar_conf["oauth_key"] = key
    click.echo("Save key {}".format(key))
    anyconfig.dump(settings.quasar_conf, settings.CONFIG_FILE, ac_parser=settings.CONFIG_FORMAT)


def oauth_print_key():
    config_check()
    click.echo("OAuth key: {}".format(settings.quasar_conf["oauth_key"]))


def config_check():
    if not os.path.isfile(settings.CONFIG_FILE):
        f = open(settings.CONFIG_FILE, "wb+")
        f.write(msgspec.json.encode(settings.quasar_conf))
        f.close()

    settings.quasar_conf = anyconfig.load(settings.CONFIG_FILE, ac_parser=settings.CONFIG_FORMAT)

    if "oauth_key" not in settings.quasar_conf:
        click.echo("Please store OAuth key with oauth save-key command.")
        click.echo("Use oauth help command to get detailed instructions.")
        exit(1)


def quasar_get_device_id_by_external_id(external_id):
    """Smart home info"""
    config_check()
    req = Request("{}/user/info".format(settings.quasar_conf["base_url"]))
    req.method = "GET"
    req.add_header('Authorization', 'Bearer {}'.format(settings.quasar_conf["oauth_key"]))
    data = decode(urlopen(req).read(), type=QuasarResponse)
    obj = data.devices

    for record in obj:
        if record.external_id == external_id:
            return record.id
    return None



import pprintpp

from yandex_quasar_cli.quasar_types.quasar_capabilities import *
from yandex_quasar_cli.quasar_types.quasar_properties import *


# https://yandex.ru/dev/dialogs/smart-home/doc/concepts/platform-protocol.html

class QuasarObject(Struct, kw_only=True):
    def pprint(self):
        click.echo(pprintpp.pformat(self), color=True)

class Status(Struct, kw_only=True):
    request_id: str
    status: str


class DeviceObject(QuasarObject, kw_only=True):
    id: str
    name: str
    aliases: list[str]
    room: Optional[str] = None
    external_id: str
    skill_id: str
    type: str
    groups: list[str]
    capabilities: list[DeviceCapabilityObject]
    properties: list[Optional[DevicePropertyObject]]

    def print_tsv(self):
        click.echo("{}\t{}\t{}\t{}".format(self.id, self.external_id, self.name, self.type))

    def print(self):
        click.echo("Id: {}\nExternal Id: {}\nName: {}\nType:{}\n".format(self.id, self.external_id, self.name,
                                                                         self.type))


class DeviceState(DeviceObject, kw_only=True):
    state: str

    def print_tsv(self):
        click.echo("{}\t{}\t{}\t{}\t{}".format(self.id, self.external_id, self.name, self.type, self.state))

    def print(self):
        click.echo(
            "Id: {}\nExternal Id: {}\nName: {}\nType: {}\nState: {}\n".format(self.id, self.external_id, self.name,
                                                                              self.type, self.state))
        for p in self.properties:
            p.print()

class RoomObject(QuasarObject, kw_only=True):
    id: str
    name: str
    household_id: str
    devices: list[str]

    def print(self):
        click.echo("{} {}".format(self.id, self.name))

    def print_tsv(self):
        click.echo("Id: {}\nName: {}\n".format(self.id, self.name))


class HouseholdObject(QuasarObject, kw_only=True):
    id: str
    name: str
    type: str


class GroupObject(QuasarObject, kw_only=True):
    id: str
    name: str
    aliases: list[str]
    type: str
    capabilities: list[Optional[GroupCapabilityObject]]
    devices: list[str]
    household_id: str

    def print_tsv(self):
        click.echo("{}\t{}".format(self.id, self.name))

    def print(self):
        click.echo("Id: {}\nName: {}\n".format(self.id, self.name))


class ScenarioObject(QuasarObject, kw_only=True):
    id: str
    name: str
    is_active: bool

    def print_tsv(self):
        click.echo("{}\t{}".format(self.id, self.name))

    def print(self):
        click.echo("Id: {}\nName: {}\n".format(self.id, self.name))


# https://yandex.ru/dev/dialogs/smart-home/doc/concepts/platform-user-info.html
class QuasarResponse(QuasarObject, kw_only=True):
    status: str
    request_id: str
    rooms: list[RoomObject]
    groups: list[GroupObject]
    devices: list[DeviceObject]
    households: list[HouseholdObject]
    scenarios: list[ScenarioObject]

    def to_dict(self):
        return {f: getattr(self, f) for f in self.__struct_fields__}


from msgspec import Struct
from typing import Optional


class CapabilityStateObject(Struct, kw_only=True):
    instance: str
    value: object


class DeviceCapabilityObject(Struct, kw_only=True):
    type: str
    reportable: Optional[bool] = None
    retrievable: bool
    parameters: object
    state: Optional[CapabilityStateObject] = None
    last_updated: float


class GroupCapabilityObject(Struct, kw_only=True):
    type: str
    retrievable: bool
    parameters: object
    state: Optional[CapabilityStateObject] = None



from msgspec import Struct
from typing import Optional
from datetime import datetime

import click


class PropertyParameterObject(Struct, kw_only=True):
    instance: str
    unit: Optional[str] = None


class PropertyStateObject(Struct, kw_only=True):
    instance: str
    value: object


class DevicePropertyObject(Struct, kw_only=True):
    type: str
    retrievable: bool
    reportable: Optional[bool] = None
    parameters: PropertyParameterObject
    state: Optional[PropertyStateObject] = None
    last_updated: float

    def print(self):
        click.echo("Type: {}".format(self.type))
        click.echo("Reportable: {}".format(self.reportable))
        click.echo("Retrievable: {}".format(self.retrievable))
        if self.parameters is not None:
            click.echo("Parameters:\n Instance: {}\n Unit: {}".format(self.parameters.instance, self.parameters.unit))
        if self.state is not None:
            click.echo("State instance: {}".format(self.state.instance))
            click.echo("State value: {}".format(self.state.value))
        click.echo("Last updated: {:%Y-%m-%d %H:%M:%S}".format(datetime.fromtimestamp(self.last_updated)))
        click.echo()



import click
import anyconfig

from yandex_quasar_cli import settings
from yandex_quasar_cli.actions import quasar_on_off, quasar_delete, quasar_get_info, quasar_get_full_info, \
    quasar_save_snapshot, quasar_compare_snapshot, config_check
from yandex_quasar_cli.actions import oauth_help, oauth_save_key, oauth_print_key


@click.group()
@click.version_option()
def cli():
    """Command line interface for Yandex Smart Home API"""


# OAuth key operations #################################################################################################
@cli.group("oauth")
def oauth():
    """OAuth key operation"""
    pass


@oauth.command(name="help")
def oauth_help_command():
    """How to get OAuth key"""
    oauth_help()


@oauth.command(name="save-key")
@click.argument(
    "key",
    required=True
)
def oauth_save_key_command(key):
    """How to get OAuth key"""
    oauth_save_key(key)


@oauth.command(name="print-key")
def oauth_print_key_command():
    """Print stored OAuth key"""
    oauth_print_key()


# Snapshots operations #################################################################################################
@cli.group("snapshot")
def snapshot():
    """Quasar state snapshots operation"""
    pass


@snapshot.command(name="save")
def snapshot_save_command():
    """Save Quasar state snapshot"""
    quasar_save_snapshot()


@snapshot.command(name="compare")
@click.option(
    "--full",
    help="Full output",
    is_flag=True,
)
@click.option(
    "--no-props-caps",
    help="No properties or capabilities compare",
    is_flag=True,
)
@click.option(
    "--external-ids",
    help="Show external ids of changed objects only",
    is_flag=True,
)
@click.option(
    "--ids",
    help="Show external and internal ids of changed objects only",
    is_flag=True,
)
def snapshot_compare_command(full, no_props_caps, external_ids, ids):
    """Compare Quasar state with saved snapshot"""
    quasar_compare_snapshot(full, no_props_caps, external_ids, ids)


# Quasar operations ####################################################################################################
@cli.group("quasar")
def quasar():
    """Quasar objects info"""
    pass


@quasar.command(name="devices")
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def quasar_devices_command(tsv):
    """Quasar devices info"""
    quasar_get_full_info("devices", tsv)


@quasar.command(name="rooms")
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def quasar_rooms_command(tsv):
    """Quasar rooms info"""
    quasar_get_full_info("rooms", tsv)


@quasar.command(name="households")
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def quasar_households_command(tsv):
    """Quasar households info"""
    quasar_get_full_info("households", tsv)


@quasar.command(name="groups")
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def quasar_groups_command(tsv):
    """Quasar groups info"""
    quasar_get_full_info("groups", tsv)


@quasar.command(name="scenarios")
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def quasar_scenarios_command(tsv):
    """Quasar scenarios info"""
    quasar_get_full_info("scenarios", tsv)


# Device operations ####################################################################################################
@cli.group("device")
def device():
    """Device operations"""
    pass


@device.command(name="info")
@click.argument(
    "device_id",
    required=True
)
@click.option(
    "--tsv",
    help="TSV out",
    is_flag=True,
)
def device_info_command(device_id, tsv):
    quasar_get_info(device_id, tsv)


@device.command(name="delete")
@click.option(
    "--ext",
    help="Use external id (if omitted use internal id)",
    is_flag=True,
)
@click.argument(
    "device_id",
    required=True
)
def device_delete_command(ext, device_id):
    quasar_delete(ext, device_id)


@device.command(name="switch-on")
@click.argument(
    "device_id",
    required=True
)
def device_switch_on_command(device_id):
    quasar_on_off(device_id, True)


@device.command(name="switch-off")
@click.argument(
    "device_id",
    required=True
)
def device_switch_off_command(device_id):
    quasar_on_off(device_id, False)



from os.path import expanduser

home = expanduser("~")

CONFIG_FILE = "{}/.yandex-quasar-cli.conf".format(home)
CONFIG_FORMAT = "json"

quasar_conf = {"base_url": "https://api.iot.yandex.net/v1.0", "snapshot": {}}




from yandex_quasar_cli.quasar_types.quasar_capabilities import *


class CapabilityObject(Struct, kw_only=True):
    type: str
    state: CapabilityStateObject


class DeviceActionObject(Struct, kw_only=True):
    id: str
    actions: list[CapabilityObject]


class ActionResultObject(Struct, kw_only=True):
    status: str
    error_code: str
    error_message: str


class StateResultObject(Struct, kw_only=True):
    instance: str
    action_result: ActionResultObject


class CapabilityActionResultObject(Struct, kw_only=True):
    type: str
    state: StateResultObject


class DeviceActionsResultObject(Struct, kw_only=True):
    id: str
    capabilities: list[CapabilityActionResultObject]


class DeviceActionRequest(Struct, kw_only=True):
    devices: list[DeviceActionObject]

class DeviceActionResponse(Struct, kw_only=True):
    request_id: str
    status: str
    devices: list[DeviceActionsResultObject]




from yandex_quasar_cli.quasar_types.quasar_capabilities import *


class DeviceCapabilityOnOff(Struct, kw_only=True):
    type: str
    state: Optional[CapabilityStateObject] = None


class DeviceCapabilityOnOffActionResult(Struct, kw_only=True):
    status: str
    error_code: str
    error_message: str


class DeviceCapabilityOnOffStateRespone(Struct, kw_only=True):
    type: str
    state: Optional[CapabilityStateObject] = None
    action_result: DeviceCapabilityOnOffActionResult



from setuptools import setup

setup(
    name='yandex-quasar-cli',
    version='0.1.2',
    packages=['yandex_quasar_cli', 'yandex_quasar_cli.quasar_types'],
    url='https://github.com/p1ne/yandex-quasar-cli',
    license='Apache-2.0',
    author='p1ne',
    author_email='',
    description='Command line interface for Yandex Smart Home API'
)
